@@ -1,21 +1,37 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
-/*
- * E-table-item.c: A GnomeCanvasItem that is a view of an ETableModel.
- *
- * Author:
- *   Christopher James Lahey <clahey@ximian.com>
- *   Miguel de Icaza (miguel@gnu.org)
- *
+/* 
+ * e-table-item.c
  * Copyright 1999, 2000, 2001, Ximian, Inc.
  *
+ * Authors:
+ *   Chris Lahey <clahey@ximian.com>
+ *   Miguel de Icaza <miguel@gnu.org>
+ *
+ * This library is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Library General Public
+ * License, version 2, as published by the Free Software Foundation.
+ *
+ * This library is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Library General Public License for more details.
+ *
+ * You should have received a copy of the GNU Library General Public
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
+ * 02111-1307, USA.
+ */
+
+/*
  * TODO:
  *   Add a border to the thing, so that focusing works properly.
- *
  */
 #include <config.h>
 
 #include "e-table-item.h"
 
+#include <X11/Xlib.h>
+
 #include <math.h>
 #include <stdio.h>
 #include <gtk/gtksignal.h>
@@ -27,14 +43,20 @@
 #include "gal/widgets/e-canvas-utils.h"
 #include "gal/util/e-util.h"
 #include <string.h>
+#include <stdlib.h>
 
 #define PARENT_OBJECT_TYPE gnome_canvas_item_get_type ()
 
 #define FOCUSED_BORDER 2
 
-#define DO_TOOLTIPS 1
-
 #define d(x)
+#define DO_TOOLTIPS
+
+#if d(!)0
+#define e_table_item_leave_edit_(x) (e_table_item_leave_edit((x)), g_print ("%s: e_table_item_leave_edit\n", __FUNCTION__))
+#else
+#define e_table_item_leave_edit_(x) (e_table_item_leave_edit((x)))
+#endif
 
 /* FIXME: Do an analysis of which cell functions are needed before
    realize and make sure that all of them are doable by all the cells
@@ -50,10 +72,11 @@ enum {
 	CLICK,
 	KEY_PRESS,
 	START_DRAG,
+	STYLE_SET,
 	LAST_SIGNAL
 };
 
-static gint eti_signals [LAST_SIGNAL] = { 0, };
+static guint eti_signals [LAST_SIGNAL] = { 0, };
 
 enum {
 	ARG_0,
@@ -71,7 +94,7 @@ enum {
 	
 	ARG_MINIMUM_WIDTH,
 	ARG_WIDTH,
-	ARG_HEIGHT,
+	ARG_HEIGHT
 };
 
 #define DOUBLE_CLICK_TIME      250
@@ -84,6 +107,8 @@ static void e_table_item_focus (ETableItem *eti, int col, int row, GdkModifierTy
 static void eti_cursor_change (ESelectionModel *selection, int row, int col, ETableItem *eti);
 static void eti_cursor_activated (ESelectionModel *selection, int row, int col, ETableItem *eti);
 static void eti_selection_change (ESelectionModel *selection, ETableItem *eti);
+static void eti_selection_row_change (ESelectionModel *selection, int row, ETableItem *eti);
+static void e_table_item_redraw_row (ETableItem *eti, int row);
 
 #define ETI_SINGLE_ROW_HEIGHT(eti) ((eti)->uniform_row_height_cache != -1 ? (eti)->uniform_row_height_cache : eti_row_height((eti), -1))
 #define ETI_MULTIPLE_ROW_HEIGHT(eti,row) ((eti)->height_cache && (eti)->height_cache[(row)] != -1 ? (eti)->height_cache[(row)] : eti_row_height((eti),(row)))
@@ -146,9 +171,62 @@ view_to_model_col(ETableItem *eti, int col)
 	return ecol ? ecol->col_idx : -1;
 }
 
+static void
+grab_cancelled (ECanvas *canvas, GnomeCanvasItem *item, gpointer data)
+{
+	ETableItem *eti = data;
+
+	eti->grab_cancelled = TRUE;
+}
+
+inline static void
+eti_grab (ETableItem *eti, guint32 time)
+{
+	GnomeCanvasItem *item = GNOME_CANVAS_ITEM (eti);
+	d(g_print ("%s: time: %d\n", __FUNCTION__, time));
+	if (eti->grabbed_count == 0) {
+		eti->gtk_grabbed = FALSE;
+		eti->grab_cancelled = FALSE;
+		if (e_canvas_item_grab(E_CANVAS (item->canvas),
+				       item,
+				       GDK_BUTTON1_MOTION_MASK | GDK_BUTTON2_MOTION_MASK | GDK_BUTTON3_MOTION_MASK
+				       | GDK_POINTER_MOTION_MASK | GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK,
+				       NULL, time,
+				       grab_cancelled,
+				       eti) != GrabSuccess) {
+			d(g_print ("%s: gtk_grab_add\n", __FUNCTION__));
+			gtk_grab_add (GTK_WIDGET (item->canvas));
+			eti->gtk_grabbed = TRUE;
+		}
+	}
+	eti->grabbed_count ++;
+}
+
+inline static void
+eti_ungrab (ETableItem *eti, guint32 time)
+{
+	GnomeCanvasItem *item = GNOME_CANVAS_ITEM (eti);
+	d(g_print ("%s: time: %d\n", __FUNCTION__, time));
+	eti->grabbed_count --;
+	if (eti->grabbed_count == 0) {
+		if (eti->grab_cancelled) {
+			eti->grab_cancelled = FALSE;
+		} else {
+			if (eti->gtk_grabbed) {
+				d(g_print ("%s: gtk_grab_remove\n", __FUNCTION__));
+				gtk_grab_remove (GTK_WIDGET (item->canvas));
+				eti->gtk_grabbed = FALSE;
+			}
+			gnome_canvas_item_ungrab(item, time);
+		}
+	}
+}
+
 inline static gboolean
 eti_editing (ETableItem *eti)
 {
+	d(g_print("%s: %s\n", __FUNCTION__, (eti->editing_col == -1) ? "false":"true"));
+	
 	if (eti->editing_col == -1)
 		return FALSE;
 	else
@@ -221,6 +299,20 @@ eti_get_cell_foreground_color (ETableItem *eti, int row, int col, gboolean selec
 	return foreground;
 }
 
+static void
+eti_free_save_state (ETableItem *eti)
+{
+	if (eti->save_row == -1 ||
+	    !eti->cell_views_realized)
+		return;
+
+	e_cell_free_state (eti->cell_views [eti->save_col], view_to_model_col(eti, eti->save_col),
+			   eti->save_col, eti->save_row, eti->save_state);
+	eti->save_row = -1;
+	eti->save_col = -1;
+	eti->save_state = NULL;
+}
+
 /*
  * During realization, we have to invoke the per-ecell realize routine
  * (On our current setup, we have one e-cell per column.
@@ -283,6 +375,8 @@ eti_unrealize_cell_views (ETableItem *eti)
 	if (eti->cell_views_realized == 0)
 		return;
 	
+	eti_free_save_state (eti);
+
 	for (i = 0; i < eti->n_cells; i++)
 		e_cell_unrealize (eti->cell_views [i]);
 	eti->cell_views_realized = 0;
@@ -293,6 +387,8 @@ eti_detach_cell_views (ETableItem *eti)
 {
 	int i;
 	
+	eti_free_save_state (eti);
+
 	for (i = 0; i < eti->n_cells; i++){
 		e_cell_kill_view (eti->cell_views [i]);
 		eti->cell_views [i] = NULL;
@@ -401,6 +497,8 @@ eti_remove_table_model (ETableItem *eti)
 
 	gtk_signal_disconnect (GTK_OBJECT (eti->table_model),
 			       eti->table_model_pre_change_id);
+	gtk_signal_disconnect (GTK_OBJECT (eti->table_model),
+			       eti->table_model_no_change_id);
 	gtk_signal_disconnect (GTK_OBJECT (eti->table_model),
 			       eti->table_model_change_id);
 	gtk_signal_disconnect (GTK_OBJECT (eti->table_model),
@@ -416,6 +514,7 @@ eti_remove_table_model (ETableItem *eti)
 		gtk_object_unref (GTK_OBJECT (eti->source_model));
 
 	eti->table_model_pre_change_id = 0;
+	eti->table_model_no_change_id = 0;
 	eti->table_model_change_id = 0;
 	eti->table_model_row_change_id = 0;
 	eti->table_model_cell_change_id = 0;
@@ -439,6 +538,8 @@ eti_remove_selection_model (ETableItem *eti)
 
 	gtk_signal_disconnect (GTK_OBJECT (eti->selection),
 			       eti->selection_change_id);
+	gtk_signal_disconnect (GTK_OBJECT (eti->selection),
+			       eti->selection_row_change_id);
 	gtk_signal_disconnect (GTK_OBJECT (eti->selection),
 			       eti->cursor_change_id);
 	gtk_signal_disconnect (GTK_OBJECT (eti->selection),
@@ -446,6 +547,7 @@ eti_remove_selection_model (ETableItem *eti)
 	gtk_object_unref (GTK_OBJECT (eti->selection));
 
 	eti->selection_change_id = 0;
+	eti->selection_row_change_id = 0;
 	eti->cursor_activated_id = 0;
 	eti->selection = NULL;
 }
@@ -545,19 +647,21 @@ height_cache_idle(ETableItem *eti)
 static void
 free_height_cache (ETableItem *eti)
 {
-	if (eti->height_cache)
-		g_free (eti->height_cache);
-	eti->height_cache = NULL;
-	eti->height_cache_idle_count = 0;
-	eti->uniform_row_height_cache = -1;
+	if (GNOME_CANVAS_ITEM_REALIZED & GTK_OBJECT_FLAGS (eti)) {
+		if (eti->height_cache)
+			g_free (eti->height_cache);
+		eti->height_cache = NULL;
+		eti->height_cache_idle_count = 0;
+		eti->uniform_row_height_cache = -1;
 
-	if (eti->uniform_row_height && eti->height_cache_idle_id != 0) {
-		g_source_remove(eti->height_cache_idle_id);
-		eti->height_cache_idle_id = 0;
+		if (eti->uniform_row_height && eti->height_cache_idle_id != 0) {
+			g_source_remove(eti->height_cache_idle_id);
+			eti->height_cache_idle_id = 0;
+		}
+
+		if ((!eti->uniform_row_height) && eti->height_cache_idle_id == 0)
+			eti->height_cache_idle_id = g_idle_add_full(G_PRIORITY_LOW, (GSourceFunc) height_cache_idle, eti, NULL);
 	}
-
-	if ((!eti->uniform_row_height) && eti->height_cache_idle_id == 0)
-		eti->height_cache_idle_id = g_idle_add_full(G_PRIORITY_LOW, (GSourceFunc) height_cache_idle, eti, NULL);
 }
 
 static void
@@ -784,16 +888,22 @@ eti_show_cursor (ETableItem *eti, int delay)
 	
 	if (!((GTK_OBJECT_FLAGS(eti) & GNOME_CANVAS_ITEM_REALIZED) && eti->cell_views_realized))
 		return;
+
+	if (eti->frozen_count > 0) {
+		eti->queue_show_cursor = TRUE;
+		return;
+	}
 	
 	gtk_object_get(GTK_OBJECT(eti->selection),
 		       "cursor_row", &cursor_row,
 		       NULL);
 
+	d(g_print ("%s: cursor row: %d\n", __FUNCTION__, cursor_row));
+
 	if (cursor_row != -1) {
 		cursor_row = model_to_view_row (eti, cursor_row);
-		d(g_print ("%s: cursor row: %d\n", __FUNCTION__, cursor_row));
 		eti_request_region_show (eti,
-					 0, cursor_row, eti->cols + 1, cursor_row + 1,
+					 0, cursor_row, eti->cols - 1, cursor_row,
 					 delay);
 	}
 }
@@ -824,6 +934,10 @@ eti_check_cursor_bounds (ETableItem *eti)
 
 	if (!((GTK_OBJECT_FLAGS(eti) & GNOME_CANVAS_ITEM_REALIZED) && eti->cell_views_realized))
 		return;
+
+	if (eti->frozen_count > 0) {
+		return;
+	}
 	
 	gtk_object_get(GTK_OBJECT(eti->selection),
 		       "cursor_row", &cursor_row,
@@ -834,16 +948,18 @@ eti_check_cursor_bounds (ETableItem *eti)
 		eti->cursor_y1 = -1;
 		eti->cursor_x2 = -1;
 		eti->cursor_y2 = -1;
-		eti->cursor_on_screen = FALSE;
+		eti->cursor_on_screen = TRUE;
 		return;
 	}
 
+	d(g_print ("%s: model cursor row: %d\n", __FUNCTION__, cursor_row));
+
 	cursor_row = model_to_view_row (eti, cursor_row);
 
 	d(g_print ("%s: cursor row: %d\n", __FUNCTION__, cursor_row));
 
 	eti_get_region (eti,
-			0, cursor_row, eti->cols + 1, cursor_row + 1,
+			0, cursor_row, eti->cols - 1, cursor_row,
 			&x1, &y1, &x2, &y2);
 	eti->cursor_x1 = x1;
 	eti->cursor_y1 = y1;
@@ -864,11 +980,12 @@ eti_maybe_show_cursor(ETableItem *eti, int delay)
 }
 
 static gboolean
-eti_idle_maybe_show_cursor_cb (gpointer data)
+eti_idle_show_cursor_cb (gpointer data)
 {
 	ETableItem *eti = data;
 	if (!GTK_OBJECT_DESTROYED (eti)) {
-		eti_maybe_show_cursor (eti, 0);
+		eti_show_cursor (eti, 0);
+		eti_check_cursor_bounds (eti);
 	}
 	gtk_object_unref (GTK_OBJECT (eti));
 	return FALSE;
@@ -877,34 +994,86 @@ eti_idle_maybe_show_cursor_cb (gpointer data)
 static void
 eti_idle_maybe_show_cursor(ETableItem *eti)
 {
-	gtk_object_ref (GTK_OBJECT (eti));
-	g_idle_add (eti_idle_maybe_show_cursor_cb, eti);
+	d(g_print ("%s: cursor on screen: %s\n", __FUNCTION__, eti->cursor_on_screen ? "TRUE" : "FALSE"));
+	if (eti->cursor_on_screen) {
+		gtk_object_ref (GTK_OBJECT (eti));
+		g_idle_add (eti_idle_show_cursor_cb, eti);
+	}
+}
+
+static void
+eti_cancel_drag_due_to_model_change (ETableItem *eti)
+{
+	if (eti->maybe_in_drag) {
+		eti->maybe_in_drag = FALSE;
+		if (!eti->maybe_did_something)
+			e_selection_model_do_something(E_SELECTION_MODEL (eti->selection), eti->drag_row, eti->drag_col, eti->drag_state);
+	}
+	if (eti->in_drag) {
+		eti->in_drag = FALSE;
+	}
+}
+
+static void
+eti_freeze (ETableItem *eti)
+{
+	eti->frozen_count ++;
+	d(g_print ("%s: %d\n", __FUNCTION__, eti->frozen_count));
+}
+
+static void
+eti_unfreeze (ETableItem *eti)
+{
+	g_return_if_fail (eti->frozen_count > 0);
+	eti->frozen_count --;
+	d(g_print ("%s: %d\n", __FUNCTION__, eti->frozen_count));
+	if (eti->frozen_count == 0 && eti->queue_show_cursor) {
+		eti_show_cursor (eti, 0);
+		eti_check_cursor_bounds (eti);
+		eti->queue_show_cursor = FALSE;
+	}
 }
 
 /*
- * Callback routine: invoked before the ETableModel has suffers a change
+ * Callback routine: invoked before the ETableModel suffers a change
  */
 static void
 eti_table_model_pre_change (ETableModel *table_model, ETableItem *eti)
 {
+	eti_cancel_drag_due_to_model_change (eti);
 	eti_check_cursor_bounds (eti);
 	if (eti_editing (eti))
-		e_table_item_leave_edit (eti);
+		e_table_item_leave_edit_(eti);
+	eti_freeze (eti);
+}
+
+/*
+ * Callback routine: invoked when the ETableModel has not suffered a change
+ */
+static void
+eti_table_model_no_change (ETableModel *table_model, ETableItem *eti)
+{
+	eti_unfreeze (eti);
 }
 
 /*
  * Callback routine: invoked when the ETableModel has suffered a change
  */
+
 static void
 eti_table_model_changed (ETableModel *table_model, ETableItem *eti)
 {
-	if (!(GTK_OBJECT_FLAGS(eti) & GNOME_CANVAS_ITEM_REALIZED))
+ 	if (!(GTK_OBJECT_FLAGS(eti) & GNOME_CANVAS_ITEM_REALIZED)) {
+		eti_unfreeze (eti);
 		return;
+	}
 
 	eti->rows = e_table_model_row_count (eti->table_model);
 
 	free_height_cache(eti);
 
+	eti_unfreeze (eti);
+
 	eti->needs_compute_height = 1;
 	e_canvas_item_request_reflow (GNOME_CANVAS_ITEM (eti));
 	eti->needs_redraw = 1;
@@ -916,36 +1085,46 @@ eti_table_model_changed (ETableModel *table_model, ETableItem *eti)
 static void
 eti_table_model_row_changed (ETableModel *table_model, int row, ETableItem *eti)
 {
-	if (!(GTK_OBJECT_FLAGS(eti) & GNOME_CANVAS_ITEM_REALIZED))
+ 	if (!(GTK_OBJECT_FLAGS(eti) & GNOME_CANVAS_ITEM_REALIZED)) {
+		eti_unfreeze (eti);
 		return;
+	}
 
 	if ((!eti->uniform_row_height) && eti->height_cache && eti->height_cache[row] != -1 && eti_row_height_real(eti, row) != eti->height_cache[row]) {
 		eti_table_model_changed (table_model, eti);
 		return;
 	}
 
-	eti_request_region_redraw (eti, 0, row, eti->cols - 1, row, 0);
+	eti_unfreeze (eti);
+
+	e_table_item_redraw_row (eti, row);
 }
 
 static void
 eti_table_model_cell_changed (ETableModel *table_model, int col, int row, ETableItem *eti)
 {
-	if (!(GTK_OBJECT_FLAGS(eti) & GNOME_CANVAS_ITEM_REALIZED))
+ 	if (!(GTK_OBJECT_FLAGS(eti) & GNOME_CANVAS_ITEM_REALIZED)) {
+		eti_unfreeze (eti);
 		return;
+	}
 
 	if ((!eti->uniform_row_height) && eti->height_cache && eti->height_cache[row] != -1 && eti_row_height_real(eti, row) != eti->height_cache[row]) {
 		eti_table_model_changed (table_model, eti);
 		return;
 	}
 
-	eti_request_region_redraw (eti, 0, row, eti->cols - 1, row, 0);
+	eti_unfreeze (eti);
+
+	e_table_item_redraw_row (eti, row);
 }
 
 static void
 eti_table_model_rows_inserted (ETableModel *table_model, int row, int count, ETableItem *eti)
 {
-	if (!(GTK_OBJECT_FLAGS(eti) & GNOME_CANVAS_ITEM_REALIZED))
+ 	if (!(GTK_OBJECT_FLAGS(eti) & GNOME_CANVAS_ITEM_REALIZED)) {
+		eti_unfreeze (eti);
 		return;
+	}
 	eti->rows = e_table_model_row_count (eti->table_model);
 
 	if (eti->height_cache) {
@@ -956,6 +1135,8 @@ eti_table_model_rows_inserted (ETableModel *table_model, int row, int count, ETa
 			eti->height_cache[i] = -1;
 	}
 
+	eti_unfreeze (eti);
+
 	eti_idle_maybe_show_cursor(eti);
 
 	eti->needs_compute_height = 1;
@@ -967,8 +1148,10 @@ eti_table_model_rows_inserted (ETableModel *table_model, int row, int count, ETa
 static void
 eti_table_model_rows_deleted (ETableModel *table_model, int row, int count, ETableItem *eti)
 {
-	if (!(GTK_OBJECT_FLAGS(eti) & GNOME_CANVAS_ITEM_REALIZED))
+ 	if (!(GTK_OBJECT_FLAGS(eti) & GNOME_CANVAS_ITEM_REALIZED)) {
+		eti_unfreeze (eti);
 		return;
+	}
 
 	g_assert (eti->rows == -1 || row + count <= eti->rows);
 
@@ -980,6 +1163,8 @@ eti_table_model_rows_deleted (ETableModel *table_model, int row, int count, ETab
 		memmove(eti->height_cache + row, eti->height_cache + row + count, (eti->rows - row) * sizeof(int));
 	}
 
+	eti_unfreeze (eti);
+
 	eti_idle_maybe_show_cursor(eti);
 
 	eti->needs_compute_height = 1;
@@ -1026,6 +1211,14 @@ e_table_item_redraw_range (ETableItem *eti,
 	eti_request_region_redraw(eti, start_col, start_row, end_col, end_row, border);
 }
 
+static void
+e_table_item_redraw_row (ETableItem *eti,
+			 int row)
+{
+	if (row != -1)
+		e_table_item_redraw_range (eti, 0, row, eti->cols - 1, row);
+}
+
 static void
 eti_add_table_model (ETableItem *eti, ETableModel *table_model)
 {
@@ -1038,6 +1231,10 @@ eti_add_table_model (ETableItem *eti, ETableModel *table_model)
 		GTK_OBJECT (table_model), "model_pre_change",
 		GTK_SIGNAL_FUNC (eti_table_model_pre_change), eti);
 
+	eti->table_model_no_change_id = gtk_signal_connect (
+		GTK_OBJECT (table_model), "model_no_change",
+		GTK_SIGNAL_FUNC (eti_table_model_no_change), eti);
+
 	eti->table_model_change_id = gtk_signal_connect (
 		GTK_OBJECT (table_model), "model_changed",
 		GTK_SIGNAL_FUNC (eti_table_model_changed), eti);
@@ -1069,7 +1266,8 @@ eti_add_table_model (ETableItem *eti, ETableModel *table_model)
 		if (eti->source_model)
 			gtk_object_ref(GTK_OBJECT(eti->source_model));
 	}
-	
+
+	eti_freeze (eti);
 	eti_table_model_changed (table_model, eti);
 }
 
@@ -1085,6 +1283,10 @@ eti_add_selection_model (ETableItem *eti, ESelectionModel *selection)
 		GTK_OBJECT (selection), "selection_changed",
 		GTK_SIGNAL_FUNC (eti_selection_change), eti);
 
+	eti->selection_row_change_id = gtk_signal_connect (
+		GTK_OBJECT (selection), "selection_row_changed",
+		GTK_SIGNAL_FUNC (eti_selection_row_change), eti);
+
 	eti->cursor_change_id = gtk_signal_connect (
 		GTK_OBJECT (selection), "cursor_changed",
 		GTK_SIGNAL_FUNC (eti_cursor_change), eti);
@@ -1191,15 +1393,18 @@ eti_destroy (GtkObject *object)
 	eti->height_cache_idle_count = 0;
 
 	e_canvas_hide_tooltip (E_CANVAS(GNOME_CANVAS_ITEM(eti)->canvas));
-	if (eti->tooltip->background)
-		gdk_color_free (eti->tooltip->background);
-	if (eti->tooltip->foreground)
-		gdk_color_free (eti->tooltip->foreground);
-	if (eti->tooltip->timer) {
-		gtk_timeout_remove (eti->tooltip->timer);
-		eti->tooltip->timer = 0;
+	if (eti->tooltip) {
+		if (eti->tooltip->background)
+			gdk_color_free (eti->tooltip->background);
+		if (eti->tooltip->foreground)
+			gdk_color_free (eti->tooltip->foreground);
+		if (eti->tooltip->timer) {
+			gtk_timeout_remove (eti->tooltip->timer);
+			eti->tooltip->timer = 0;
+		}
+		g_free (eti->tooltip);
 	}
-	g_free (eti->tooltip);
+	eti->tooltip = NULL;
 
 	if (GTK_OBJECT_CLASS (eti_parent_class)->destroy)
 		(*GTK_OBJECT_CLASS (eti_parent_class)->destroy) (object);
@@ -1335,6 +1540,10 @@ eti_init (GnomeCanvasItem *item)
 	eti->width                     = 0;
 	eti->minimum_width             = 0;
 
+	eti->save_col                  = -1;
+	eti->save_row                  = -1;
+	eti->save_state                = NULL;
+
 	eti->click_count               = 0;
 
 	eti->height_cache              = NULL;
@@ -1351,10 +1560,13 @@ eti_init (GnomeCanvasItem *item)
 	eti->cursor_mode               = E_CURSOR_SIMPLE;
 
 	eti->selection_change_id       = 0;
+	eti->selection_row_change_id   = 0;
 	eti->cursor_change_id          = 0;
 	eti->cursor_activated_id       = 0;
 	eti->selection                 = NULL;
 
+	eti->old_cursor_row            = -1;
+
 	eti->needs_redraw              = 0;
 	eti->needs_compute_height      = 0;
 
@@ -1368,13 +1580,16 @@ eti_init (GnomeCanvasItem *item)
 
 	eti->maybe_did_something       = TRUE;
 
-	eti->grabbed_col               = -1;
-	eti->grabbed_row               = -1;
+	eti->grabbed_count             = 0;
+	eti->gtk_grabbed               = 0;
 
 	eti->in_drag                   = 0;
 	eti->maybe_in_drag             = 0;
 	eti->grabbed                   = 0;
 
+	eti->grabbed_col               = -1;
+	eti->grabbed_row               = -1;
+
 	eti->cursor_on_screen          = FALSE;
  	eti->cursor_x1                 = -1;
 	eti->cursor_y1                 = -1;
@@ -1383,6 +1598,9 @@ eti_init (GnomeCanvasItem *item)
 
 	eti->rows                      = -1;
 
+	eti->frozen_count              = 0;
+	eti->queue_show_cursor         = FALSE;
+
 	e_canvas_item_set_reflow_callback (GNOME_CANVAS_ITEM (eti), eti_reflow);
 }
 
@@ -1447,6 +1665,18 @@ eti_realize (GnomeCanvasItem *item)
 
 	free_height_cache(eti);
 
+	if (item->canvas->focused_item == NULL && eti->selection) {
+		int row;
+
+		row = e_selection_model_cursor_row (E_SELECTION_MODEL (eti->selection));
+		row = model_to_view_row(eti, row);
+		if (row != -1) {
+			e_canvas_item_grab_focus (item, FALSE);
+			eti_show_cursor (eti, 0);
+			eti_check_cursor_bounds (eti);
+		}
+	}
+
 	eti->needs_compute_height = 1;
 	eti->needs_compute_width = 1;
 	e_canvas_item_request_reflow (GNOME_CANVAS_ITEM (eti));
@@ -1460,12 +1690,37 @@ eti_unrealize (GnomeCanvasItem *item)
 	ETableItem *eti = E_TABLE_ITEM (item);
 
 	if (eti->grabbed) {
-		gtk_grab_remove (GTK_WIDGET (item->canvas));
 		eti->grabbed = FALSE;
+		d(g_print ("%s: eti_ungrab\n", __FUNCTION__));
+		eti_ungrab (eti, -1);
 	}
 
 	if (eti_editing (eti))
-		e_table_item_leave_edit (eti);
+		e_table_item_leave_edit_(eti);
+
+	if (eti->height_cache_idle_id) {
+		g_source_remove(eti->height_cache_idle_id);
+		eti->height_cache_idle_id = 0;
+	}
+
+	if (eti->height_cache)
+		g_free (eti->height_cache);
+	eti->height_cache = NULL;
+	eti->height_cache_idle_count = 0;
+
+	e_canvas_hide_tooltip (E_CANVAS(GNOME_CANVAS_ITEM(eti)->canvas));
+	if (eti->tooltip->background) {
+		gdk_color_free (eti->tooltip->background);
+		eti->tooltip->background = NULL;
+	}
+	if (eti->tooltip->foreground) {
+		gdk_color_free (eti->tooltip->foreground);
+		eti->tooltip->foreground = NULL;
+	}
+	if (eti->tooltip->timer) {
+		gtk_timeout_remove (eti->tooltip->timer);
+		eti->tooltip->timer = 0;
+	}
 
 	gdk_gc_unref (eti->fill_gc);
 	eti->fill_gc = NULL;
@@ -1629,6 +1884,7 @@ eti_draw (GnomeCanvasItem *item, GdkDrawable *drawable, int x, int y, int width,
 			ETableCol *ecol = e_table_header_get_column (eti->header, col);
 			ECellView *ecell_view = eti->cell_views [col];
 			gboolean col_selected = selected;
+			gboolean cursor = FALSE;
 			ECellFlags flags;
 			gboolean free_background;
 			GdkColor *background;
@@ -1636,8 +1892,10 @@ eti_draw (GnomeCanvasItem *item, GdkDrawable *drawable, int x, int y, int width,
 			switch (eti->cursor_mode) {
 			case E_CURSOR_SIMPLE:
 			case E_CURSOR_SPREADSHEET:
-				if (cursor_col == ecol->col_idx && cursor_row == view_to_model_row(eti, row))
+				if (cursor_col == ecol->col_idx && cursor_row == view_to_model_row(eti, row)) {
 					col_selected = !col_selected;
+					cursor = TRUE;
+				}
 				break;
 			case E_CURSOR_LINE:
 				/* Nothing */
@@ -1655,6 +1913,8 @@ eti_draw (GnomeCanvasItem *item, GdkDrawable *drawable, int x, int y, int width,
 
 			flags = col_selected ? E_CELL_SELECTED : 0;
 			flags |= GTK_WIDGET_HAS_FOCUS(canvas) ? E_CELL_FOCUSED : 0;
+			flags |= cursor ? E_CELL_CURSOR : 0;
+
 			switch (ecol->justification) {
 			case GTK_JUSTIFY_LEFT:
 				flags |= E_CELL_JUSTIFY_LEFT;
@@ -1816,7 +2076,7 @@ find_cell (ETableItem *eti, double x, double y, int *view_col_res, int *view_row
 static void
 eti_cursor_move (ETableItem *eti, gint row, gint column)
 {
-	e_table_item_leave_edit (eti);
+	e_table_item_leave_edit_(eti);
 	e_table_item_focus (eti, view_to_model_col(eti, column), view_to_model_row(eti, row), 0);
 }
 
@@ -1918,13 +2178,15 @@ eti_e_cell_event     (ETableItem *item, ECellView *ecell_view, GdkEvent *event,
 	ret_val = e_cell_event (ecell_view, event, model_col, view_col, row, flags, &actions);
 
 	if (actions & E_CELL_GRAB) {
-		gnome_canvas_item_grab(GNOME_CANVAS_ITEM(item), GDK_POINTER_MOTION_MASK | GDK_BUTTON_RELEASE_MASK, NULL, time);
+		d(g_print ("%s: eti_grab\n", __FUNCTION__));
+		eti_grab (item, time);
 		item->grabbed_col = view_col;
 		item->grabbed_row = row;
 	}
 
 	if (actions & E_CELL_UNGRAB) {
-		gnome_canvas_item_ungrab(GNOME_CANVAS_ITEM(item), time);
+		d(g_print ("%s: eti_ungrab\n", __FUNCTION__));
+		eti_ungrab (item, time);
 		item->grabbed_col = -1;
 		item->grabbed_row = -1;
 	}
@@ -1939,6 +2201,9 @@ eti_event (GnomeCanvasItem *item, GdkEvent *e)
 	ETableItem *eti = E_TABLE_ITEM (item);
 	ECellView *ecell_view;
 	gint return_val = TRUE;
+#if d(!)0
+	gboolean leave = FALSE;
+#endif
 
 	switch (e->type){
 	case GDK_BUTTON_PRESS: {
@@ -1948,6 +2213,9 @@ eti_event (GnomeCanvasItem *item, GdkEvent *e)
 		int col, row;
 		gint cursor_row, cursor_col;
 		gint new_cursor_row, new_cursor_col;
+		ECellFlags flags = 0;
+
+		d(g_print("%s: GDK_BUTTON_PRESS received, button %d\n", __FUNCTION__, e->button.button));
 
 		if (eti->tooltip->timer) {
 			gtk_timeout_remove (eti->tooltip->timer);
@@ -1965,7 +2233,7 @@ eti_event (GnomeCanvasItem *item, GdkEvent *e)
 
 			if (!find_cell (eti, realx, realy, &col, &row, &x1, &y1)) {
 				if (eti_editing (eti))
-					e_table_item_leave_edit (eti);
+					e_table_item_leave_edit_(eti);
 				return TRUE;
 			}
 
@@ -1974,11 +2242,21 @@ eti_event (GnomeCanvasItem *item, GdkEvent *e)
 			button.x = x1;
 			button.y = y1;
 
-			return_val = eti_e_cell_event (eti, ecell_view, (GdkEvent *) &button, button.time, view_to_model_col(eti, col), col, row, 0);
+			gtk_object_get(GTK_OBJECT(eti->selection),
+				       "cursor_row", &cursor_row,
+				       "cursor_col", &cursor_col,
+				       NULL);
+
+			if (cursor_col == view_to_model_col (eti, col) && cursor_row == view_to_model_row(eti, row)) {
+				flags = E_CELL_CURSOR;
+			} else {
+				flags = 0;
+			}
+
+			return_val = eti_e_cell_event (eti, ecell_view, (GdkEvent *) &button, button.time, view_to_model_col(eti, col), col, row, flags);
 			if (return_val)
 				return TRUE;
 
-			return_val = FALSE;
 			gtk_signal_emit (GTK_OBJECT (eti), eti_signals [CLICK],
 					 row, view_to_model_col(eti, col), &button, &return_val);
 
@@ -2015,13 +2293,15 @@ eti_event (GnomeCanvasItem *item, GdkEvent *e)
 
 				if (eti_editing (eti)) {
 					return_val = eti_e_cell_event (eti, ecell_view, (GdkEvent *) &button, button.time, 
-								       view_to_model_col(eti, col), col, row, E_CELL_EDITING);
+								       view_to_model_col(eti, col), col, row, E_CELL_EDITING | E_CELL_CURSOR);
 					if (return_val)
 						return TRUE;
 				}
 			}
 
 			if (e->button.button == 1) {
+				return_val = TRUE;
+
 				eti->maybe_in_drag = TRUE;
 				eti->drag_row      = new_cursor_row;
 				eti->drag_col      = new_cursor_col;
@@ -2029,14 +2309,10 @@ eti_event (GnomeCanvasItem *item, GdkEvent *e)
 				eti->drag_y        = realy;
 				eti->drag_state    = e->button.state;
 				eti->grabbed       = TRUE;
-				if (!gnome_canvas_item_grab(item,
-							    (1 << (4 + e->button.button)) | GDK_POINTER_MOTION_MASK | GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK, 
-							    NULL, e->button.time))
-					gtk_grab_add (GTK_WIDGET (item->canvas));
+				d(g_print ("%s: eti_grab\n", __FUNCTION__));
+				eti_grab (eti, e->button.time);
 			}
 
-			d(g_print("Single click\n"));
-
 			break;
 		case 3:
 			e_canvas_item_grab_focus(GNOME_CANVAS_ITEM(eti), TRUE);
@@ -2065,12 +2341,15 @@ eti_event (GnomeCanvasItem *item, GdkEvent *e)
 		int col, row;
 		gint cursor_row, cursor_col;
 
-		if (e->button.button == 1) {
-			if (eti->grabbed) {
-				gtk_grab_remove (GTK_WIDGET (item->canvas));
-				gnome_canvas_item_ungrab(item, e->button.time);
-			}
+		d(g_print("%s: GDK_BUTTON_RELEASE received, button %d\n", __FUNCTION__, e->button.button));
+
+		if (eti->grabbed) {
+			d(g_print ("%s: eti_ungrab\n", __FUNCTION__));
+			eti_ungrab (eti, e->button.time);
 			eti->grabbed = FALSE;
+		}
+
+		if (e->button.button == 1) {
 			if (eti->maybe_in_drag) {
 				eti->maybe_in_drag = FALSE;
 				if (!eti->maybe_did_something)
@@ -2091,6 +2370,13 @@ eti_event (GnomeCanvasItem *item, GdkEvent *e)
 		case 2:
 
 			gnome_canvas_item_w2i (item, &e->button.x, &e->button.y);
+#if d(!)0
+			{
+				gboolean cell_found = find_cell (eti, e->button.x, e->button.y, &col, &row, &x1, &y1);
+				g_print("%s: find_cell(%f, %f) = %s(%d, %d, %f, %f)\n", __FUNCTION__, e->button.x, e->button.y,
+					cell_found?"true":"false", col, row, x1, y1);
+			}
+#endif
 
 			if (!find_cell (eti, e->button.x, e->button.y, &col, &row, &x1, &y1))
 				return TRUE;
@@ -2100,8 +2386,16 @@ eti_event (GnomeCanvasItem *item, GdkEvent *e)
 				       "cursor_col", &cursor_col,
 				       NULL);
 
+
+			d(g_print("%s: GDK_BUTTON_RELEASE received, button %d, line: %d\n"
+				  "eti_editing: %s, row:%d:%d, col:%d:%d\n", __FUNCTION__, e->button.button, __LINE__,
+				  eti_editing(eti)?"true":"false", cursor_row, view_to_model_row(eti, row), cursor_col, view_to_model_col(eti, col)));
+
 			if (eti_editing (eti) && cursor_row == view_to_model_row(eti, row) && cursor_col == view_to_model_col(eti, col)){
 
+				d(g_print("%s: GDK_BUTTON_RELEASE received, button %d, line: %d\n", __FUNCTION__, e->button.button, __LINE__))
+;
+
 				ecell_view = eti->cell_views [col];
 
 				/*
@@ -2111,9 +2405,8 @@ eti_event (GnomeCanvasItem *item, GdkEvent *e)
 				e->button.y = y1;
 
 				return_val = eti_e_cell_event (eti, ecell_view, e, e->button.time,
-							       view_to_model_col(eti, col), col, row, E_CELL_EDITING);
+							       view_to_model_col(eti, col), col, row, E_CELL_EDITING | E_CELL_CURSOR);
 			}
-			d(g_print ("Release\n"));
 			break;
 		case 3:
 			e_selection_model_right_click_up(E_SELECTION_MODEL (eti->selection));
@@ -2133,6 +2426,9 @@ eti_event (GnomeCanvasItem *item, GdkEvent *e)
 #if 0
 		double x1, y1;
 #endif
+
+		d(g_print("%s: GDK_2BUTTON_PRESS received, button %d\n", __FUNCTION__, e->button.button));
+
 		if (e->button.button == 5 ||
 		    e->button.button == 4)
 			return FALSE;
@@ -2164,10 +2460,26 @@ eti_event (GnomeCanvasItem *item, GdkEvent *e)
 			button.y = y1;
 #endif
 
+			if (e->button.button == 1) {
+				if (eti->maybe_in_drag) {
+					eti->maybe_in_drag = FALSE;
+					if (!eti->maybe_did_something)
+						e_selection_model_do_something(E_SELECTION_MODEL (eti->selection), eti->drag_row, eti->drag_col, eti->drag_state);
+				}
+				if (eti->in_drag) {
+					eti->in_drag = FALSE;
+				}
+			}
+
+			if (eti->grabbed) {
+				d(g_print ("%s: eti_ungrab\n", __FUNCTION__));
+				eti_ungrab (eti, e->button.time);
+				eti->grabbed = FALSE;
+			}
+
 			if (model_row != -1 && model_col != -1) {
 				gtk_signal_emit (GTK_OBJECT (eti), eti_signals [DOUBLE_CLICK],
 						 model_row, model_col, e);
-				d(g_print("Double click\n"));
 			}
 		}
 		break;
@@ -2226,7 +2538,7 @@ eti_event (GnomeCanvasItem *item, GdkEvent *e)
 			e->motion.y = y1;
 
 			return_val = eti_e_cell_event (eti, ecell_view, e, e->motion.time,
-						       view_to_model_col(eti, col), col, row, E_CELL_EDITING);
+						       view_to_model_col(eti, col), col, row, E_CELL_EDITING | E_CELL_CURSOR);
 		}
 		break;
 	}
@@ -2235,6 +2547,8 @@ eti_event (GnomeCanvasItem *item, GdkEvent *e)
 		gint cursor_row, cursor_col;
 		gint handled = TRUE;
 
+		d(g_print("%s: GDK_KEY_PRESS received, keyval: %d\n", __FUNCTION__, (int) e->key.keyval));
+
 		gtk_object_get(GTK_OBJECT(eti->selection),
 			       "cursor_row", &cursor_row,
 			       "cursor_col", &cursor_col,
@@ -2343,12 +2657,12 @@ eti_event (GnomeCanvasItem *item, GdkEvent *e)
 		case GDK_ISO_Enter:
 		case GDK_3270_Enter:
 			if (eti_editing (eti)){
-				e_table_item_leave_edit (eti);
+				e_table_item_leave_edit_(eti);
 #if 0
 				ecell_view = eti->cell_views [eti->editing_col];
 				return_val = eti_e_cell_event (eti, ecell_view, e, e->key.time,
 							       view_to_model_col(eti, eti->editing_col),
-							       eti->editing_col, eti->editing_row, E_CELL_EDITING);
+							       eti->editing_col, eti->editing_row, E_CELL_EDITING | E_CELL_CURSOR);
 #endif
 			}
 			gtk_signal_emit (GTK_OBJECT (eti), eti_signals [KEY_PRESS],
@@ -2401,7 +2715,7 @@ eti_event (GnomeCanvasItem *item, GdkEvent *e)
 					ecell_view = eti->cell_views [eti->editing_col];
 					return_val = eti_e_cell_event (eti, ecell_view, e, e->key.time,
 								       view_to_model_col(eti, eti->editing_col),
-								       eti->editing_col, eti->editing_row, E_CELL_EDITING);
+								       eti->editing_col, eti->editing_row, E_CELL_EDITING | E_CELL_CURSOR);
 					if (!return_val)
 						e_selection_model_key_press(E_SELECTION_MODEL (eti->selection), (GdkEventKey *) e);
 				}
@@ -2415,6 +2729,8 @@ eti_event (GnomeCanvasItem *item, GdkEvent *e)
 	case GDK_KEY_RELEASE: {
 		gint cursor_row, cursor_col;
 
+		d(g_print("%s: GDK_KEY_RELEASE received, keyval: %d\n", __FUNCTION__, (int) e->key.keyval));
+
 		gtk_object_get(GTK_OBJECT(eti->selection),
 			       "cursor_row", &cursor_row,
 			       "cursor_col", &cursor_col,
@@ -2427,30 +2743,77 @@ eti_event (GnomeCanvasItem *item, GdkEvent *e)
 			ecell_view = eti->cell_views [eti->editing_col];
 			return_val = eti_e_cell_event (eti, ecell_view, e, e->key.time,
 						       view_to_model_col(eti, eti->editing_col),
-						       eti->editing_col, eti->editing_row, E_CELL_EDITING);
+						       eti->editing_col, eti->editing_row, E_CELL_EDITING | E_CELL_CURSOR);
 		}
 		break;
 	}
 
 	case GDK_LEAVE_NOTIFY:
+		d(leave = TRUE);
 	case GDK_ENTER_NOTIFY:
+		d(g_print("%s: %s received\n", __FUNCTION__, leave ? "GDK_LEAVE_NOTIFY" : "GDK_ENTER_NOTIFY"));
 		if (eti->tooltip->timer > 0)
 			gtk_timeout_remove (eti->tooltip->timer);
 		eti->tooltip->timer = 0;
 		break;
 
 	case GDK_FOCUS_CHANGE:
-		if (! e->focus_change.in) {
-			if (eti_editing (eti))
-				e_table_item_leave_edit (eti);
+		d(g_print("%s: GDK_FOCUS_CHANGE received, %s\n", __FUNCTION__, e->focus_change.in ? "in": "out"));
+		if (e->focus_change.in) {
+			if (eti->save_row != -1 &&
+			    eti->save_col != -1 &&
+			    !eti_editing (eti) &&
+			    e_table_model_is_cell_editable(eti->table_model, view_to_model_col (eti, eti->save_col), eti->save_row)) { 
+				e_table_item_enter_edit (eti, eti->save_col, eti->save_row);
+				e_cell_load_state (eti->cell_views [eti->editing_col], view_to_model_col(eti, eti->save_col),
+						   eti->save_col, eti->save_row, eti->edit_ctx, eti->save_state);
+				eti_free_save_state (eti);
+			}
+		} else {
+			if (eti_editing (eti)) {
+				eti_free_save_state (eti);
+
+				eti->save_row   = eti->editing_row;
+				eti->save_col   = eti->editing_col;
+				eti->save_state = e_cell_save_state (eti->cell_views [eti->editing_col], view_to_model_col(eti, eti->editing_col),
+								     eti->editing_col, eti->editing_row, eti->edit_ctx);
+				e_table_item_leave_edit_(eti);
+			}
 		}
 
 	default:
 		return_val = FALSE;
 	}
+	/* d(g_print("%s: returning: %s\n", __FUNCTION__, return_val?"true":"false"));*/
+
 	return return_val;
+}
+
+static void
+eti_style_set (ETableItem *eti, GtkStyle *previous_style)
+{
+ 	if (!(GTK_OBJECT_FLAGS(eti) & GNOME_CANVAS_ITEM_REALIZED))
+		return;
+
+	if (eti->cell_views_realized) {
+		int i;
+		int n_cells = eti->n_cells;
+
+		for (i = 0; i < n_cells; i++) {
+			e_cell_style_set (eti->cell_views[i], previous_style);
+		}
 	}
 
+	eti->needs_compute_height = 1;
+	e_canvas_item_request_reflow (GNOME_CANVAS_ITEM (eti));
+	eti->needs_redraw = 1;
+	gnome_canvas_item_request_update (GNOME_CANVAS_ITEM (eti));
+
+	free_height_cache (eti);
+
+	eti_idle_maybe_show_cursor(eti);
+}
+
 static void
 eti_class_init (GtkObjectClass *object_class)
 {
@@ -2477,6 +2840,7 @@ eti_class_init (GtkObjectClass *object_class)
 	eti_class->click            = NULL;
 	eti_class->key_press        = NULL;
 	eti_class->start_drag       = NULL;
+	eti_class->style_set        = eti_style_set;
 
 	gtk_object_add_arg_type ("ETableItem::ETableHeader", E_TABLE_HEADER_TYPE,
 				 GTK_ARG_WRITABLE, ARG_TABLE_HEADER);
@@ -2564,6 +2928,14 @@ eti_class_init (GtkObjectClass *object_class)
 				e_marshal_INT__INT_INT_POINTER,
 				GTK_TYPE_INT, 3, GTK_TYPE_INT, GTK_TYPE_INT, GTK_TYPE_GDK_EVENT);
 
+	eti_signals [STYLE_SET] =
+		gtk_signal_new ("style_set",
+				GTK_RUN_LAST,
+				E_OBJECT_CLASS_TYPE (object_class),
+				GTK_SIGNAL_OFFSET (ETableItemClass, style_set),
+				gtk_marshal_NONE__POINTER,
+				GTK_TYPE_NONE, 1, GTK_TYPE_STYLE);
+
 	E_OBJECT_CLASS_ADD_SIGNALS (object_class, eti_signals, LAST_SIGNAL);
 
 }
@@ -2661,8 +3033,12 @@ eti_cursor_change (ESelectionModel *selection, int row, int col, ETableItem *eti
 	view_row = model_to_view_row(eti, row);
 	view_col = model_to_view_col(eti, col);
 	
+	if (eti->old_cursor_row != -1 && view_row != eti->old_cursor_row)
+		e_table_item_redraw_row (eti, eti->old_cursor_row);
+
 	if (view_row == -1 || view_col == -1) {
-		e_table_item_leave_edit (eti);
+		e_table_item_leave_edit_(eti);
+		eti->old_cursor_row = -1;
 		return;
 	}
 
@@ -2676,11 +3052,13 @@ eti_cursor_change (ESelectionModel *selection, int row, int col, ETableItem *eti
 
 	e_canvas_item_grab_focus(GNOME_CANVAS_ITEM(eti), FALSE);
 	if (eti_editing(eti))
-		e_table_item_leave_edit (eti);
+		e_table_item_leave_edit_(eti);
 	gtk_signal_emit (GTK_OBJECT (eti), eti_signals [CURSOR_CHANGE],
 			 view_row);
-	eti->needs_redraw = TRUE;
-	gnome_canvas_item_request_update(GNOME_CANVAS_ITEM(eti));
+
+	e_table_item_redraw_row (eti, view_row);
+
+	eti->old_cursor_row = view_row;
 }
 
 static void
@@ -2695,21 +3073,19 @@ eti_cursor_activated (ESelectionModel *selection, int row, int col, ETableItem *
 	view_row = model_to_view_row(eti, row);
 	view_col = model_to_view_col(eti, col);
 	
-	if (view_row == -1 || view_col == -1) {
-		e_table_item_leave_edit (eti);
-		return;
-	}
-
-	if (! e_table_model_has_change_pending (eti->table_model)) {
-		if (!eti->in_key_press) {
-			eti_show_cursor(eti, DOUBLE_CLICK_TIME + 10);
-		} else {
-			eti_show_cursor(eti, 0);
+	if (view_row != -1 && view_col != -1) {
+		if (! e_table_model_has_change_pending (eti->table_model)) {
+			if (!eti->in_key_press) {
+				eti_show_cursor(eti, DOUBLE_CLICK_TIME + 10);
+			} else {
+				eti_show_cursor(eti, 0);
+			}
+			eti_check_cursor_bounds (eti);
 		}
 	}
 
 	if (eti_editing(eti))
-		e_table_item_leave_edit (eti);
+		e_table_item_leave_edit_(eti);
 	gtk_signal_emit (GTK_OBJECT (eti), eti_signals [CURSOR_ACTIVATED],
 			 view_row);
 }
@@ -2724,6 +3100,17 @@ eti_selection_change (ESelectionModel *selection, ETableItem *eti)
 	gnome_canvas_item_request_update(GNOME_CANVAS_ITEM(eti));
 }
 
+static void
+eti_selection_row_change (ESelectionModel *selection, int row, ETableItem *eti)
+{
+	if (!(GTK_OBJECT_FLAGS(eti) & GNOME_CANVAS_ITEM_REALIZED))
+		return;
+
+	if (!eti->needs_redraw) {
+		e_table_item_redraw_row (eti, model_to_view_row(eti, row));
+	}
+}
+
 
 /** 
  * e_table_item_enter_edit
@@ -2740,8 +3127,10 @@ e_table_item_enter_edit (ETableItem *eti, int col, int row)
 	g_return_if_fail (eti != NULL);
 	g_return_if_fail (E_IS_TABLE_ITEM (eti));
 	
+	d(g_print("%s: %d, %d, eti_editing() = %s\n", __FUNCTION__, col, row, eti_editing(eti)?"true":"false"));
+
 	if (eti_editing (eti))
-		e_table_item_leave_edit(eti);
+		e_table_item_leave_edit_(eti);
 
 	eti->editing_col = col;
 	eti->editing_row = row;
@@ -2750,7 +3139,7 @@ e_table_item_enter_edit (ETableItem *eti, int col, int row)
 }
 
 /** 
- * e_table_item_leave_edit
+ * e_table_item_leave_edit_
  * @eti: %ETableItem which will stop being edited
  *
  * This routine stops the given %ETableItem from editing.
@@ -2764,6 +3153,8 @@ e_table_item_leave_edit (ETableItem *eti)
 	g_return_if_fail (eti != NULL);
 	g_return_if_fail (E_IS_TABLE_ITEM (eti));
 
+	d(g_print("%s: eti_editing() = %s\n", __FUNCTION__, eti_editing(eti)?"true":"false"));
+
 	if (!eti_editing (eti))
 		return;
 
@@ -2807,7 +3198,7 @@ e_table_item_compute_location (ETableItem        *eti,
 	eti->grabbed_row = -1;
 
 	if (!find_cell (eti, *x, *y, col, row, NULL, NULL)) {
-		*y -= eti_get_height(eti);
+		*y -= eti->height;
 	}
 
 	eti->grabbed_row = grabbed_row;
@@ -2850,7 +3241,7 @@ e_table_item_calculate_print_widths (ETableHeader *eth, gdouble width)
 	double extra;
 	double expansion;
 	int last_resizable = -1;
-	gdouble scale = 300.0L / 70.0L;
+	gdouble scale = 1.0L;
 	gdouble *widths = g_new(gdouble, e_table_header_count(eth));
 	/* - 1 to account for the last pixel border. */
 	extra = width - 1;
