@@ -1,25 +1,29 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /* 
- * e-xml-utils.c
- * Copyright (C) 2000  Helix Code, Inc.
- * Author: Chris Lahey <clahey@helixcode.com>
+ * e-util.c
+ * Copyright 2000, 2001, Ximian, Inc.
+ *
+ * Authors:
+ *   Chris Lahey <clahey@ximian.com>
  *
  * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU General Public License as
- * published by the Free Software Foundation; either version 2 of the
- * License, or (at your option) any later version.
+ * modify it under the terms of the GNU Library General Public
+ * License, version 2, as published by the Free Software Foundation.
  *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * This library is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * General Public License for more details.
+ * Library General Public License for more details.
  *
- * You should have received a copy of the GNU General Public
- * License along with this library; if not, write to the
- * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
- * Boston, MA 02111-1307, USA.
+ * You should have received a copy of the GNU Library General Public
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
+ * 02111-1307, USA.
  */
 
+#include <config.h>
+#include "e-util.h"
+
 #include <glib.h>
 #include <gtk/gtkobject.h>
 #include <errno.h>
@@ -32,8 +36,10 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <time.h>
+#include <libgnome/gnome-defs.h>
+#include <libgnome/gnome-util.h>
+#include <math.h>
 
-#include "e-util.h"
 #if 0
 #include <libgnomevfs/gnome-vfs.h>
 #endif
@@ -195,7 +201,37 @@ e_write_file(const char *filename, const char *data, int flags)
 			}
 		}
 	}
-	close(fd);
+	if (close(fd) != 0) {
+		return errno;
+	}
+	return 0;
+}
+
+gint
+e_write_file_mkstemp(char *filename, const char *data)
+{
+	int fd;
+	int length = strlen(data);
+	int bytes;
+	fd = mkstemp (filename);
+	if (fd == -1)
+		return errno;
+	while (length > 0) {
+		bytes = write(fd, data, length);
+		if (bytes > 0) {
+			length -= bytes;
+			data += bytes;
+		} else {
+			if (errno != EINTR && errno != EAGAIN) {
+				int save_errno = errno;
+				close(fd);
+				return save_errno;
+			}
+		}
+	}
+	if (close(fd) != 0) {
+		return errno;
+	}
 	return 0;
 }
 
@@ -215,13 +251,22 @@ e_mkdir_hier(const char *path, mode_t mode)
 {
 	char *copy, *p;
 
-	p = copy = g_strdup (path);
+	if (path[0] == '/') {
+		p = copy = g_strdup (path);
+	} else {
+		gchar *current_dir = g_get_current_dir();
+		p = copy = g_concat_dir_and_file (current_dir, path);
+	}
+
 	do {
 		p = strchr (p + 1, '/');
 		if (p)
 			*p = '\0';
-		if (access (copy, F_OK) == -1) {
-			if (mkdir (copy, mode) == -1) {
+		if (mkdir (copy, mode) == -1) {
+			switch (errno) {
+			case EEXIST:
+				break;
+			default:
 				g_free (copy);
 				return -1;
 			}
@@ -617,6 +662,22 @@ e_marshal_NONE__INT_POINTER_INT_POINTER (GtkObject * object, GtkSignalFunc func,
 	    GTK_VALUE_INT (args[2]), GTK_VALUE_POINTER (args[3]), func_data);
 }	
 
+typedef void (*GtkSignal_NONE__POINTER_POINTER_POINTER_POINTER) (GtkObject *,
+								 gpointer, gpointer, gpointer, gpointer,
+								 gpointer);
+
+void
+e_marshal_NONE__POINTER_POINTER_POINTER_POINTER (GtkObject         *object,
+						 GtkSignalFunc      func,
+						 gpointer           func_data,
+						 GtkArg            *args)
+{
+  GtkSignal_NONE__POINTER_POINTER_POINTER_POINTER rfunc;
+  rfunc = (GtkSignal_NONE__POINTER_POINTER_POINTER_POINTER) func;
+  (*rfunc) (object, GTK_VALUE_POINTER (args[0]), GTK_VALUE_POINTER (args[1]),
+	    GTK_VALUE_POINTER (args[2]), GTK_VALUE_POINTER (args[3]), func_data);
+}
+
 typedef int (*GtkSignal_INT__POINTER_POINTER) (GtkObject *,
 					       gpointer, gpointer, 
 					       gpointer user_data);
@@ -704,10 +765,31 @@ e_marshal_INT__POINTER_POINTER_POINTER_POINTER_POINTER (GtkObject *object,
 				GTK_VALUE_POINTER (args[1]), 
 				GTK_VALUE_POINTER (args[2]),
 				GTK_VALUE_POINTER (args[3]),
-				GTK_VALUE_POINTER (args[3]),
+				GTK_VALUE_POINTER (args[4]),
 				func_data);
 }
 
+typedef void (*GtkSignal_NONE__POINTER_POINTER_POINTER_BOOL) (GtkObject *,
+							      gpointer, gpointer, gpointer, gboolean,
+							      gpointer user_data);
+
+void
+e_marshal_NONE__POINTER_POINTER_POINTER_BOOL (GtkObject *object,
+					      GtkSignalFunc func,
+					      gpointer func_data,
+					      GtkArg *args)
+{
+	GtkSignal_NONE__POINTER_POINTER_POINTER_BOOL rfunc;
+
+	rfunc = (GtkSignal_NONE__POINTER_POINTER_POINTER_BOOL) func;
+
+	(*rfunc) (object,
+		  GTK_VALUE_POINTER (args[0]),
+		  GTK_VALUE_POINTER (args[1]), 
+		  GTK_VALUE_POINTER (args[2]),
+		  GTK_VALUE_BOOL (args[3]),
+		  func_data);
+}
 
 void
 e_marshal_NONE__POINTER_INT_INT_INT (GtkObject       *object,
@@ -746,6 +828,47 @@ e_marshal_INT__OBJECT_POINTER (GtkObject *object,
 				func_data);
 }
 
+typedef void (*GtkSignal_NONE__DOUBLE) (GtkObject *,
+				       gdouble,
+				       gpointer user_data);
+void
+e_marshal_NONE__DOUBLE (GtkObject *object,
+			GtkSignalFunc func,
+			gpointer func_data,
+			GtkArg *args)
+{
+	GtkSignal_NONE__DOUBLE rfunc;
+
+	rfunc = (GtkSignal_NONE__DOUBLE) func;
+
+	(*rfunc) (object,
+		  GTK_VALUE_DOUBLE (args[0]),
+		  func_data);
+}
+
+typedef gboolean (*GtkSignal_BOOL__STRING_INT) (GtkObject *,
+						char *,
+						gint,
+						gpointer user_data);
+
+void
+e_marshal_BOOL__STRING_INT (GtkObject *object,
+			    GtkSignalFunc func,
+			    gpointer func_data,
+			    GtkArg *args)
+{
+	GtkSignal_BOOL__STRING_INT rfunc;
+	gboolean *return_val;
+
+	rfunc = (GtkSignal_BOOL__STRING_INT) func;
+	return_val = GTK_RETLOC_BOOL (args[2]);
+
+	*return_val = (*rfunc) (object,
+				GTK_VALUE_STRING  (args[0]),
+				GTK_VALUE_INT (args[1]),
+				func_data);
+}
+
 gchar**
 e_strsplit (const gchar *string,
 	    const gchar *delimiter,
@@ -823,6 +946,7 @@ e_strstrcase (const gchar *haystack, const gchar *needle)
 	return NULL;
 }
 
+/* This only makes a filename safe for usage as a filename.  It still may have shell meta-characters in it. */
 void
 e_filename_make_safe (gchar *string)
 {
@@ -908,10 +1032,97 @@ e_format_number (gint number)
 	}
 }
 
+static gchar *
+do_format_number_as_float (double number)
+{
+	GList *iterator, *list = NULL;
+	struct lconv *locality;
+	gint char_length = 0;
+	gint group_count = 0;
+	guchar *grouping;
+	int last_count = 3;
+	int divider;
+	char *value;
+	char *value_iterator;
+	double fractional;
+
+	locality = localeconv();
+	grouping = locality->grouping;
+	while (number >= 1.0) {
+		char *group;
+		switch (*grouping) {
+		default:
+			last_count = *grouping;
+			grouping++;
+			/* Fall through */
+		case 0:
+			divider = epow10(last_count);
+			number /= divider;
+			fractional = modf (number, &number);
+			fractional *= divider;
+			fractional = floor (fractional);
+
+			if (number >= 1.0) {
+				group = g_strdup_printf("%0*d", last_count, (int) fractional);
+			} else {
+				group = g_strdup_printf("%d", (int) fractional);
+			}
+			break;
+		case CHAR_MAX:
+			divider = epow10(last_count);
+			number /= divider;
+			fractional = modf (number, &number);
+			fractional *= divider;
+			fractional = floor (fractional);
+
+			while (number >= 1.0) {
+				group = g_strdup_printf("%0*d", last_count, (int) fractional);
+
+				char_length += strlen(group);
+				list = g_list_prepend(list, group);
+				group_count ++;
+
+				divider = epow10(last_count);
+				number /= divider;
+				fractional = modf (number, &number);
+				fractional *= divider;
+				fractional = floor (fractional);
+			}
+
+			group = g_strdup_printf("%d", (int) fractional);
+			break;
+		}
+		char_length += strlen(group);
+		list = g_list_prepend(list, group);
+		group_count ++;
+	}
+
+	if (list) {
+		value = g_new(char, 1 + char_length + (group_count - 1) * strlen(locality->thousands_sep));
+
+		iterator = list;
+		value_iterator = value;
+
+		strcpy(value_iterator, iterator->data);
+		value_iterator += strlen(iterator->data);
+		for (iterator = iterator->next; iterator; iterator = iterator->next) {
+			strcpy(value_iterator, locality->thousands_sep);
+			value_iterator += strlen(locality->thousands_sep);
+
+			strcpy(value_iterator, iterator->data);
+			value_iterator += strlen(iterator->data);
+		}
+		e_free_string_list (list);
+		return value;
+	} else {
+		return g_strdup("0");
+	}
+}
+
 gchar *
 e_format_number_float (gfloat number)
 {
-	gint            int_part;
+	gfloat          int_part;
 	gint            fraction;
 	struct lconv   *locality;
 	gchar          *str_intpart;
@@ -921,8 +1132,8 @@ e_format_number_float (gfloat number)
 
 	locality = localeconv();
 	
-	int_part = (int) number;
-	str_intpart = e_format_number (int_part);
+	int_part = floor (number);
+	str_intpart = do_format_number_as_float ((double) int_part);
 
 	if (!strcmp(locality->mon_decimal_point, "")) {
 		decimal_point = ".";
@@ -935,8 +1146,7 @@ e_format_number_float (gfloat number)
 
 	if (fraction == 0) {
 		str_fraction = g_strdup ("00");
-	}
-	else {
+	} else {
 		str_fraction = g_strdup_printf ("%02d", fraction);
 	}
 
@@ -951,63 +1161,11 @@ e_format_number_float (gfloat number)
 gboolean
 e_create_directory (gchar *directory)
 {
-	gchar *full_name;
-	gchar *position;
-	gchar *current_dir = g_get_current_dir();
-	struct stat info;
-	gboolean return_value = TRUE;
-
-	if (directory[0] == '/') {
-		full_name = g_malloc0 (strlen (directory) + 1);
-		strcpy (full_name, directory);
-	} else {
-		full_name = g_malloc0 (strlen (directory) + strlen (current_dir) + 2);
-		sprintf (full_name, "%s/%s", current_dir, directory);
-	}
-
-	if ((position = strrchr (full_name, '/')) == full_name) {
-		if (stat (full_name, &info)) {
-			switch (errno) {
-			case ENOENT:
-				if (mkdir (full_name, 0777)) {
-					switch (errno) {
-					default:
-						return_value = FALSE;
-						break;
-					}
-				}
-				break;
-			default:
-				return_value = FALSE;
-				break;
-			}
-		}
-	} else {
-		*position = 0;
-		e_create_directory (full_name);
-		*position = '/';
-		if (stat (full_name, &info)) {
-			switch (errno) {
-			case ENOENT:
-				if (mkdir (full_name, 0777)) {
-					switch (errno) {
-					default:
-						return_value = FALSE;
-						break;
-					}
-				}
-				break;
-			default:
-				return_value = FALSE;
-				break;
-			}
-		}
-	}
-
-	g_free (current_dir);
-	g_free (full_name);
-
-	return (return_value);
+	gint ret_val = e_mkdir_hier (directory, 0777);
+	if (ret_val == -1)
+		return FALSE;
+	else
+		return TRUE;
 }
 
 
@@ -1115,6 +1273,34 @@ e_sort (void             *base,
 #endif
 }
 
+size_t e_strftime(char *s, size_t max, const char *fmt, const struct tm *tm)
+{
+#ifdef HAVE_LKSTRFTIME
+	return strftime(s, max, fmt, tm);
+#else
+	char *c, *ffmt, *ff;
+	size_t ret;
+
+	ffmt = g_strdup(fmt);
+	ff = ffmt;
+	while ((c = strstr(ff, "%l")) != NULL) {
+		c[1] = 'I';
+		ff = c;
+	}
+
+	ff = fmt;
+	while ((c = strstr(ff, "%k")) != NULL) {
+		c[1] = 'H';
+		ff = c;
+	}
+
+	ret = strftime(s, max, ffmt, tm);
+	g_free(ffmt);
+	return ret;
+#endif
+}
+
+
 /**
  * Function to do a last minute fixup of the AM/PM stuff if the locale
  * and gettext haven't done it right. Most English speaking countries
@@ -1142,17 +1328,17 @@ size_t e_strftime_fix_am_pm(char *s, size_t max, const char *fmt, const struct t
 
 	if (strstr(fmt, "%p")==NULL && strstr(fmt, "%P")==NULL) {
 		/* No AM/PM involved - can use the fmt string directly */
-		ret=strftime(s, max, fmt, tm);
+		ret=e_strftime(s, max, fmt, tm);
 	} else {
 		/* Get the AM/PM symbol from the locale */
-		strftime (buf, 10, "%p", tm);
+		e_strftime (buf, 10, "%p", tm);
 
 		if (buf[0]) {
 			/**
 			 * AM/PM have been defined in the locale
 			 * so we can use the fmt string directly
 			 **/
-			ret=strftime(s, max, fmt, tm);
+			ret=e_strftime(s, max, fmt, tm);
 		} else {
 			/**
 			 * No AM/PM defined by locale
@@ -1169,7 +1355,7 @@ size_t e_strftime_fix_am_pm(char *s, size_t max, const char *fmt, const struct t
 			for (sp=ffmt; (sp=strstr(sp, "%I")); sp++) {
 				sp[1]='H';
 			}
-			ret=strftime(s, max, ffmt, tm);
+			ret=e_strftime(s, max, ffmt, tm);
 			g_free(ffmt);
 		}
 	}
@@ -1453,3 +1639,29 @@ e_strdup_append_strings (gchar *first_string, ...)
 
 	return buffer;
 }
+
+gchar **
+e_strdupv (const gchar **str_array)
+{
+	if (str_array) {
+		gint i;
+		gchar **retval;
+
+		i = 0;
+		while (str_array[i])
+			i++;
+          
+		retval = g_new (gchar*, i + 1);
+
+		i = 0;
+		while (str_array[i]) {
+			retval[i] = g_strdup (str_array[i]);
+			i++;
+		}
+		retval[i] = NULL;
+
+		return retval;
+	} else {
+		return NULL;
+	}
+}
