@@ -1,17 +1,28 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
-#define _E_FONT_C_
-
-/*
- * e-font
+/* 
+ * e-font.c - Temporary wrappers around GdkFonts to get unicode displaying
+ * Copyright 2000, 2001, Ximian, Inc.
  *
- * Temporary wrappers around GdkFonts to get unicode displaying
+ * Authors:
+ *   Lauris Kaplinski <lauris@ximian.com>
  *
- * Author: Lauris Kaplinski <lauris@helixcode.com>
+ * This library is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Library General Public
+ * License, version 2, as published by the Free Software Foundation.
  *
- * Copyright (C) 2000 Helix Code, Inc.
+ * This library is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Library General Public License for more details.
  *
+ * You should have received a copy of the GNU Library General Public
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
+ * 02111-1307, USA.
  */
 
+#define _E_FONT_C_
+
 #include <config.h>
 #include <stdlib.h>
 #include <string.h>
@@ -23,6 +34,10 @@
 #include "e-unicode.h"
 #include "gal/util/e-iconv.h"
 
+#ifdef HAVE_ALLOCA_H
+#include <alloca.h>
+#endif
+
 #ifdef HAVE_CODESET
 #include <langinfo.h>
 #endif
@@ -169,13 +184,50 @@ set_nbsp_zero_width_flag (EFont *efont)
 }
 
 /*
- * Creates EFont from existing GdkFont.
- * 1. Use cache
- * 2. Analyze, whether we need font or fontset. For fontset use e_font_from_gdk_fontset
- * 3. Try, if we can simply use iso-10646 variant of font
- * 4. Determine bold or light variant
+ * Simplified version used for gtkhtml library to reduce resources usage
  */
 
+EFont *
+e_font_from_gdk_font_gtkhtml (GdkFont *gdk_font)
+{
+	EFont *font;
+	XFontStruct *xfs;
+
+#ifdef E_FONT_VERBOSE
+	if (e_font_verbose) {
+		g_print ("(e_font_from_gdk_font_gtkhtml): ");
+		e_font_print_gdk_font_name (gdk_font);
+	}
+#endif
+
+	g_return_val_if_fail (gdk_font != NULL, NULL);
+
+	font = g_new (EFont, 1);
+
+	xfs = GDK_FONT_XFONT (gdk_font);
+
+	font->refcount = 1;
+	gdk_font_ref (gdk_font);
+	gdk_font_ref (gdk_font);
+	font->font = gdk_font;
+	font->bold = gdk_font;
+	font->twobyte = (gdk_font->type == GDK_FONT_FONTSET || ((xfs->min_byte1 != 0) || (xfs->max_byte1 != 0)));
+	font->to = e_iconv_to_gdk_font (font->font);
+	font->from = e_iconv_from_gdk_font (font->font);
+	set_nbsp_zero_width_flag (font);
+
+	return font;
+
+}
+
+ /*
+  * Creates EFont from existing GdkFont.
+  * 1. Use cache
+  * 2. Analyze, whether we need font or fontset. For fontset use e_font_from_gdk_fontset
+  * 3. Try, if we can simply use iso-10646 variant of font
+  * 4. Determine bold or light variant
+  */
+
 EFont *
 e_font_from_gdk_font (GdkFont *gdkfont)
 {
@@ -712,7 +764,7 @@ replace_nbsp_with_spaces (const guchar *orig, gint *bytes)
 static gint
 e_font_to_native (EFont *font, gchar *native, const gchar *utf, gint bytes)
 {
-	char *ib;
+	const char *ib;
 	char *ob;
 	size_t ibl, obl;
 	gint rv;
@@ -723,13 +775,13 @@ e_font_to_native (EFont *font, gchar *native, const gchar *utf, gint bytes)
 	if (font->to == (iconv_t) -1)
 		rv = no_conv_wrapper (font, native, utf, bytes);
 	else {
-		ib = (char *) utf;
+		ib = utf;
 		ibl = bytes;
 		ob = native;
 		obl = bytes * 4;
 
 		while (ibl > 0) {
-			iconv (font->to, &ib, &ibl, &ob, &obl);
+			e_iconv (font->to, &ib, &ibl, &ob, &obl);
 			if (ibl > 0) {
 				ib = g_utf8_next_char (ib);
 
